# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/5 
# @Function:
import os
from pathlib import Path

import tensorrt as trt


def check_shape(shape: dict):
    assert shape, "请指定shape，例如：shape={0}".format({"input_ids": [(10, 128), (10, 128), (10, 128)],
                                                  "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                                                  "attention_mask": [(10, 128), (10, 128), (10, 128)]})
    for name in shape:
        assert len(shape.get(name)) == 3, "shape必须包含minShapes，optShapes，maxShapes三个，例如：shape={0}，但是输入为：{1}".format(
            {"input_ids": [(10, 128), (10, 128), (10, 128)]}, shape.get(name))


def convert_trt_by_command(onnx_file_path: Path, save_trt_path: Path, shape: dict, use_fp16=False):
    """
    采用trtexec工具将onnx转换成trt
    :param onnx_file_path: Path(r"model.onnx")
    :param save_trt_path: Path(r"model.trt")
    :param shape:  {"input_ids": [(10, 128), (10, 128), (10, 128)],
                   "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                   "attention_mask": [(10, 128), (10, 128), (10, 128)]}
    :param use_fp16: 使用fp16量化精度，注意：windows下转换会有问题，原因未知，理论上fp16文件时fp32文件大小的一半
    :return:
    convert_trt_by_command(onnx_file_path=Path(r"model.onnx"), save_trt_path=model.trt"),
                       shape={"input_ids": [(10, 128), (10, 128), (10, 128)],
                              "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                              "attention_mask": [(10, 128), (10, 128), (10, 128)]})
    """
    """
    --noTF32 禁用TF32精度
    --fp16   FP16量化精度
    --int8   INT8量化精度
    --best   FP32+FP16+INT8同时使用，找一个速度最快的
    --calib=xxx  指定int8校准缓存文件
    --minShapes  动态Shape指定(--minShapes=input0:1x3x256x256,input1:1x3x128x128)
    --optShapes  动态Shape指定(--optShapes=input0:1x3x256x256,input1:1x3x128x128)
    --maxShapes  动态Shape指定(--maxShapes=input0:1x3x256x256,input1:1x3x128x128)
    --inputIOFormats  指定模型输入精度与数据排布格式，默认fp32:chw(--inputIOFormats=fp16:chw)
    --outputIOFormats 指定模型输输出精度与数据排布格式，默认fp32:chw(--outputIOFormats=fp16:chw)
    --memPoolSize   内存池大小(--memPoolSize =workspace:1024.5,dlaSRAM:256)
    --profilingVerbosity  打印信息的详细程度layer_names_only|detailed|none(--profilingVerbosity=detailed)
    --explicitBatch  与固化onnx的batch size一致
    --useCudaGraph 
    """
    check_shape(shape)
    min_shapes = [shape.get(name)[0] for name in shape]
    opt_shapes = [shape.get(name)[1] for name in shape]
    max_shapes = [shape.get(name)[2] for name in shape]
    names = list(shape.keys())
    min_shapes_str = ""
    opt_shapes_str = ""
    max_shapes_str = ""
    for i in range(0, len(names)):
        min_shapes_str += "{0}:{1},".format(names[i], "x".join([str(item) for item in min_shapes[i]]))
        opt_shapes_str += "{0}:{1},".format(names[i], "x".join([str(item) for item in opt_shapes[i]]))
        max_shapes_str += "{0}:{1},".format(names[i], "x".join([str(item) for item in max_shapes[i]]))
    command = "trtexec " \
              "--onnx={0} " \
              "--saveEngine={1}" \
              "--minShapes={2} " \
              "--optShapes={3} " \
              "--maxShapes={4} ".format(onnx_file_path, save_trt_path, min_shapes_str, opt_shapes_str, max_shapes_str)
    if use_fp16:
        command += "--fp16"
    os.system(command)
    print("done.")


def convert_trt(onnx_file_path: Path, save_trt_path: Path, shape: dict, max_batch_size: int = 1, use_fp16=False):
    """
    采用OnnxParser将onnx转换成trt【推荐使用】
    :param onnx_file_path:
    :param save_trt_path:
    :param shape:  {"input_ids": [(10, 128), (10, 128), (10, 128)],
                   "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                   "attention_mask": [(10, 128), (10, 128), (10, 128)]}
    :param max_batch_size:
    :param use_fp16: 使用fp16量化精度，注意：windows下转换会有问题，原因未知，理论上fp16文件时fp32文件大小的一半
    :return:
    """

    def GiB(val):
        return val * 1 << 30

    check_shape(shape)
    assert max_batch_size <= 8, "trt推理时最大支持的batch_size=8，但是输入为：{0}".format(max_batch_size)
    g_logger = trt.Logger(trt.Logger.WARNING)
    explicit_batch = 1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    with trt.Builder(g_logger) as builder, builder.create_network(explicit_batch) as network, \
            trt.OnnxParser(network, g_logger) as parser:
        builder.max_batch_size = max_batch_size
        config = builder.create_builder_config()
        config.max_workspace_size = GiB(2)
        if use_fp16:
            config.set_flag(trt.BuilderFlag.FP16)
            config.set_flag(trt.BuilderFlag.STRICT_TYPES)
        print('Loading ONNX file from path {}...'.format(onnx_file_path))
        with open(onnx_file_path, 'rb') as model:
            print('Beginning ONNX file parsing')
            parser.parse(model.read())
        print('Completed parsing of ONNX file')
        print('Building an engine from file {}; this may take a while...'.format(onnx_file_path))
        profile = builder.create_optimization_profile()
        for name in shape:
            profile.set_shape(name, *shape.get(name))
        config.add_optimization_profile(profile)
        engine = builder.build_engine(network, config)
        assert engine
        print("Completed creating Engine")
        with open(save_trt_path, "wb") as f:
            f.write(engine.serialize())
        print("done.")


if __name__ == "__main__":
    print(trt.__version__)
    # convert_trt_by_command(onnx_file_path=Path(r"F:\torch\onnx\model.onnx"),
    #                        save_trt_path=Path(r"F:\torch\onnx\model.trt"),
    #                        shape={"input_ids": [(10, 128), (10, 128), (10, 128)],
    #                               "token_type_ids": [(10, 128), (10, 128), (10, 128)],
    #                               "attention_mask": [(10, 128), (10, 128), (10, 128)]})
    convert_trt(onnx_file_path=Path(r"F:\torch\onnx\model.onnx"),
                save_trt_path=Path(r"F:\torch\onnx\model_32.trt"),
                use_fp16=True,
                shape={"input_ids": [(10, 128), (10, 128), (50, 128)],
                       "token_type_ids": [(10, 128), (10, 128), (50, 128)],
                       "attention_mask": [(10, 128), (10, 128), (50, 128)]})
    pass




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2022/9/23
# @Function: 逻辑回归
import os
import jieba
import warnings
import pandas as pd
import numpy as np
from pathlib import Path
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.preprocessing import OneHotEncoder
from zhousflib.ml.feature_vector import FeatureVector, TypeFeatureVector
from zhousflib.ml.model_base import BaseModel


class LRClassifier(BaseModel):

    def __init__(self, model_file, vector_pickle, vector_type=TypeFeatureVector.TYPE_TF_IDF,
                 importance_files=None, is_slim_vector=False):
        """

        :param model_file: 模型文件
        :param vector_pickle: 向量文件
        :param vector_type: 特征向量表示器
        :param importance_files: 不重要特征文件
        :param is_slim_vector: True时使用importance_files
        """
        self.model_file = model_file
        if vector_pickle is None:
            warnings.warn("vector_pickle is None.")
        if model_file is None:
            warnings.warn("model_file is None.")
        self.vector = FeatureVector(default_pkl=vector_pickle, vector_type=vector_type)
        model = self.load_model(model_file) if os.path.exists(model_file) else None
        self.params = {}
        self.words_not_importance = self.read_words_not_importance(importance_files) if is_slim_vector else []
        super().__init__(model, model_file)

    def build_feature(self, excel_files: list, one_hot_class=False):
        print("构建数据特征...")
        dfs = []
        for excel_file in excel_files:
            df = pd.read_excel(excel_file, header=0)
            dfs.append(df)
        data_frame = pd.concat(dfs)
        data_frame = data_frame.fillna("")
        class_attr_name = "type"
        if one_hot_class:
            # 将类别进行OneHotEncoder编码
            class_value_list = np.arange(3).tolist()
            hot_en = OneHotEncoder(categories=[class_value_list])
            y = hot_en.fit_transform(data_frame[class_attr_name].values.reshape(-1, 1)).toarray()
        else:
            y = data_frame[class_attr_name]
        features = []
        data = []
        for index, row in data_frame.iterrows():
            row_list = list(row)
            name = row_list[0]
            label_gt = row_list[1]
            full_str = row_list[2]
            full_str = full_str.replace(" ", "")
            full_str = jieba.lcut(full_str)
            # 删除指定特征
            for word in full_str:
                if word in self.words_not_importance:
                    full_str.remove(word)
            data.append(" ".join(full_str))
            vec_tf = self.vector.transform([" ".join(full_str)])
            features.append((name, label_gt, vec_tf))
        # 数据向量化
        vec_tf = self.vector.transform(data)
        x = vec_tf.toarray()
        return features, x, y, self.vector.vocabulary

    def get_top(self, predict_prob_result, top):
        prob_dict = {}
        predict_result = predict_prob_result[0]
        for i in range(0, len(predict_result)):
            prob_dict[i] = predict_result[i]
        result = sorted(prob_dict.items(), key=lambda p: p[1], reverse=True)
        top = len(result) if top > len(result) else top
        return result[:top]

    def predict(self, string, top=1):
        """
        预测
        :param string:
        :param top:
        :return: label_pre, confidence
        """
        data = []
        full_str = string.replace(" ", "")
        full_str = jieba.lcut(full_str)
        # 删除指定特征
        for word in full_str:
            if word in self.words_not_importance:
                full_str.remove(word)
        data.append(" ".join(full_str))
        prob = self.model.predict_proba(self.vector.transform(data))
        return self.get_top(prob, top=top)[0]

    def search_best_params(self, excel_file):
        x_train, x_test, y_train, y_test = self.load_data(excel_file, one_hot_class=False, shuffle=True, test_size=0.2)
        params = {'C': [50, 100],
                  'max_iter': [100],
                  }
        print("search_best_params...")
        clf = GridSearchCV(LogisticRegression(), param_grid=params, cv=10)
        clf.fit(x_train, y_train)
        print(clf.best_params_)
        self.params.update(clf.best_params_)

    def show_feature_importance(self):
        """
        展示特征的重要性并保存文件
        https://scikit-learn.org/stable/auto_examples/ensemble/plot_forest_importances.html
        :return:
        """
        vec_word = []
        # 将特征字典排序
        word_vec = sorted(self.vector.vocabulary.items(), key=lambda x: x[1], reverse=False)
        for word in word_vec:
            w, index = word
            vec_word.append(w)
        # print(model.classes_)
        for i in range(0, len(self.model.coef_)):
            word_import = {}
            for j in range(0, len(self.model.coef_[i])):
                word_import[vec_word[j]] = self.model.coef_[i][j]
            # 根据特征重要性倒叙
            word_import = sorted(word_import.items(), key=lambda x: x[1], reverse=True)
            write_data = []
            for j in word_import:
                write_data.append("{0} {1}".format(j[0], j[1]))
            importance_file = Path.cwd().joinpath("importance_lr_{0}.txt".format(i))
            importance_file.write_text("\n".join(write_data))
        print("done.")

    def get_model(self):
        return LogisticRegression(C=20, max_iter=100, multi_class="ovr", solver="sag", random_state=50)


if __name__ == "__main__":
    excel_files_ = [r"/Users/zhousf/workspace/GOVT_OCR/dispatcher/训练集.xlsx", "/Users/zhousf/workspace/GOVT_OCR/dispatcher/测试集.xlsx"]
    classifier = LRClassifier(model_file="/Users/zhousf/workspace/GOVT_OCR/dispatcher/classification/algorithm_ml/model/lr_tfidf.h5",
                              vector_pickle="/Users/zhousf/workspace/GOVT_OCR/dispatcher/classification/algorithm_ml/model/vector_tfidf.pickle",
                              vector_type=TypeFeatureVector.TYPE_TF_IDF,
                              importance_files=[], is_slim_vector=False)
    # --------- 搜索最佳参数
    # classifier.search_best_params(excel_file=excel_file_)
    # --------- 训练
    # classifier.train(excel_files=excel_files_, one_hot_class=False, shuffle=True, test_size=0.2)
    # --------- 预测
    txt = "'建筑施工企业项目负责人安全生产考核合格证书姓名：范志强证件号码：130821199109213313760999（00）企业名称北京天狼建筑工程有限公司岗位名称项目负责人证书编号：京建安B（20170147607有效期至：2020年12月31日本电子证书中中住房和城乡会核发。本证书表明持证人已通过建筑施工全生目负成绩合格。证机关：发证日期2017年19日查询网址：zjw.beijing.gov.cn制证日期实时数据，扫码验证2020年09月23日'"
    print(classifier.predict(txt))
    # --------- 测试集评估
    # classifier.val_test(excel_file=r"/Users/zhousf/workspace/GOVT_OCR/dispatcher/测试集.xlsx", classes=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])
    # --------- 展示特征的重要程度
    # classifier.show_feature_importance()
    # --------- 卡方校验
    # classifier.chi_square_check(excel_file=excel_file_, one_hot_class=False, k=500)
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/19 
# @Function:
from backend import Backend


class BackendOriginal(Backend):
    def __init__(self, *args, **kwargs):
        self.model_dir = kwargs.get("model_dir")
        self.device_id = kwargs.get("device_id")
        self.model = None
        self.backend = "original"
        super().__init__(*args)

    def build(self, **kwargs):
        module = kwargs.get("module")
        self.pop(kwargs, "module")
        self.model = module

    def inference(self, inputs_list: list):
        return self.model(inputs_list)


# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2022/9/21
# @Function: 机器学习模型基类
import abc
import re
import joblib
from prettytable import PrettyTable
from pathlib import Path
from sklearn import metrics
from sklearn.linear_model import LogisticRegression
from sklearn.feature_selection import chi2
from sklearn.feature_selection import SelectKBest
from sklearn.model_selection import train_test_split


class BaseModel(metaclass=abc.ABCMeta):

    def __init__(self, model, model_file):
        self.model = model
        self.model_file = model_file
        pass

    @staticmethod
    def save_model(model, model_file):
        """
        保存模型
        :param model: 模型
        :param model_file: 保存模型的文件路径
        :return:
        """
        joblib.dump(model, model_file)
        print("save model: {0}".format(model_file))

    @staticmethod
    def load_model(model_file):
        """
        加载模型
        :param model_file: 模型的文件路径
        :return: 模型
        """
        print("load model: {0}".format(model_file))
        return joblib.load(model_file)

    @abc.abstractmethod
    def build_feature(self, excel_files: list, one_hot_class=False):
        """
        构建特征
        :param excel_files:
        :param one_hot_class:
        :return:
        """
        return []

    @abc.abstractmethod
    def get_top(self, predict_prob_result, top):
        """
        获取结果top
        :param predict_prob_result:
        :param top:
        :return:
        """
        return []

    @abc.abstractmethod
    def get_model(self):
        """
        获取模型实例
        :return:
        """
        return LogisticRegression()

    def load_data(self, excel_files: list, one_hot_class=False, test_size=0.2, shuffle=True):
        """
        加载数据
        :param excel_files:
        :param one_hot_class: 是否对类别进行one hot编码
        :param test_size:
        :param shuffle:
        :return:
        """
        # 数据向量化
        features, x, y, vocabulary, data = self.build_feature(excel_files, one_hot_class)
        # 拆分训练集、验证集
        return train_test_split(x, y, test_size=test_size, shuffle=shuffle, random_state=2022)

    def train(self, excel_files: list, one_hot_class=False, shuffle=True, test_size=0.2):
        """
        训练
        :param excel_files:
        :param one_hot_class:
        :param shuffle:
        :param test_size:
        :return:
        """
        # 构建训练数据
        x_train, x_test, y_train, y_test = self.load_data(excel_files, one_hot_class=one_hot_class, shuffle=shuffle,
                                                          test_size=test_size)
        print("训练...")
        model = self.get_model()
        model.fit(x_train, y_train)
        y_pre = model.predict(x_test)
        accuracy = metrics.accuracy_score(y_test, y_pre)
        print("accuracy={0:.6f}".format(accuracy))
        # 保存模型
        self.save_model(model, self.model_file)

    def val_test(self, excel_file, classes, confidence_threshold=0):
        """
        评估测试集
        :param excel_file:
        :param classes:
        :param confidence_threshold: 置信度阈值
        :return:
        """
        statistics_dict = {}
        features, x, y, vocabulary, texts = self.build_feature(excel_file)
        for i, feature in enumerate(features):
            name, label_gt, vec = feature
            # 预测
            prob = self.model.predict_proba(vec)
            top_one = self.get_top(prob, top=1)
            label_pre, confidence = top_one[0]
            if label_gt != label_pre:
                print("标注>", classes[label_gt], "预测>", classes[label_pre],label_pre, confidence, name)
                print(texts[i])
                print(top_one)
                print("---------")
            if label_gt not in statistics_dict:
                statistics_dict[label_gt] = {"TP": 0, "LABEL_COUNT": 0, "PRE_COUNT": 0}
            statistics_dict[label_gt]["LABEL_COUNT"] += 1
            if label_pre not in statistics_dict:
                statistics_dict[label_pre] = {"TP": 0, "LABEL_COUNT": 0, "PRE_COUNT": 0}
            statistics_dict[label_pre]["PRE_COUNT"] += 1
            # 置信度阈值
            if confidence < confidence_threshold:
                continue
            if label_pre == label_gt:
                statistics_dict[label_pre]["TP"] += 1
        self.show_statistics(statistics_dict, classes)

    @staticmethod
    def read_words_not_importance(importance_files):
        """
        读取特征重要程度文件，返回不重要特征list
        :param importance_files:
        :return:
        """
        # 构建不重要特征集合
        words_not_importance = []
        if importance_files is not None:
            for importance_file in importance_files:
                importance_file = Path(importance_file)
                txt = importance_file.read_text()
                for line in txt.split("\n"):
                    word, importance = line.split(" ")
                    # importance等于0既不重要
                    if float(importance) == 0:
                        words_not_importance.append(word)
        return words_not_importance

    def chi_square_check(self, excel_file, one_hot_class, k=10):
        """
        卡方校验
        :param excel_file:
        :param one_hot_class:
        :param k: Number of top features to select. int or "all", default=10
        :return:
        result = model.fit_transform(x, y)
        print(model.scores_)
        print(model.pvalues_)
        print(result)
        """
        # 数据向量化
        features, x, y, vocabulary = self.build_feature(excel_file, one_hot_class)
        vocabulary = {vocabulary.get(k): k for k in vocabulary}
        print("卡方校验...")
        # 选择k个最佳特征
        model = SelectKBest(chi2, k=k)
        model_fit = model.fit(x, y)
        best_feature_name = []
        for index in model_fit.get_support(indices=True):
            best_feature_name.append(vocabulary.get(index))
        print(best_feature_name)

    @staticmethod
    def show_statistics(statistics_dict: dict, classes: list):
        print("显示统计结果")
        table = PrettyTable(
            ["label", "f1", "precision", "recall", "TP", "label count(GT)", "label count(PRE)"])
        table.align["label"] = "l"
        table.float_format = ".4"
        table.sortby = "label"
        recall_all = 0
        precision_all = 0
        label_count = 0
        recall_mean = 0
        precision_mean = 0
        items_count = 0
        tp_count = 0
        pre_count = 0
        gt_count = 0
        for class_id, value in statistics_dict.items():
            items_count += 1
            print("=>", classes[int(class_id)], value)
            if value['LABEL_COUNT'] == 0:
                recall = 0
            else:
                recall = value["TP"] / value['LABEL_COUNT']
            tp_count += value["TP"]
            pre_count += value["PRE_COUNT"]
            gt_count += value['LABEL_COUNT']
            all_count = value["LABEL_COUNT"] if value["PRE_COUNT"] == 0 else value["PRE_COUNT"]
            precision = value["TP"] / all_count
            recall_all += recall
            precision_all += precision
            if recall + precision == 0: continue
            f1 = 2 * recall * precision / (recall + precision)
            table.add_row(
                [classes[int(class_id)], f1, precision, recall, value["TP"], value['LABEL_COUNT'],
                 value["PRE_COUNT"]])
            recall_mean += recall * all_count
            precision_mean += precision * all_count
            label_count += all_count
        recall_mean /= label_count
        precision_mean /= label_count
        precision = precision_all / items_count
        recall = recall_all / items_count
        f1 = 2 * recall * precision / (recall + precision)
        table.add_row(["ALL", f1, precision, recall, tp_count, gt_count, pre_count])
        mf1 = 2 * recall_mean * precision_mean / (recall_mean + precision_mean)
        table.add_row(["ALL_MEAN", mf1, precision_mean, recall_mean, tp_count, gt_count, pre_count])
        print(table)



# -*- coding:utf-8 -*-
# Author:  zhousf
# Date:    2022-09-24
# Description: 聚类
import numpy as np
import matplotlib.pyplot as mp
from sklearn.cluster import KMeans


def kmeans(data: np.ndarray, n_clusters=2, show=False):
    km = KMeans(n_clusters=n_clusters)
    km.fit(data)
    labels_ = km.labels_
    centers_ = km.cluster_centers_
    if show:
        colors = ['red', 'blue', 'green', 'black']
        for i in range(len(centers_)):
            x = centers_[i]
            mp.scatter(x[0], x[1], c=colors[i], marker='x', s=99)
        for x, l in zip(data, labels_):
            mp.scatter(x[0], x[1], c=colors[l])
        mp.show()
    return centers_, labels_


if __name__ == "__main__":
    txt = ['经营范围', '许可项目：', '：住宿服务；', '餐饮服务；', '；酒类经营；食品经', '住所', '海南省海口市美兰区灵山镇琼山大道', '营；小食杂；', '；高危险性体育运动（游泳）；住宅室内装', '166号', '饰装修：旅游业务。（依法须经批准的项目，经相关部', '门批准后方可开展经营活动，具体经营项目以相关部门', '批准文件或许可证件为准）一般项目：酒店管理；物业', '管理；餐饮管理；会议及展览服务；', '；住房租赁；非居住', '房地产租赁；汽车租赁；', '；家政服务；', '停车场服务；', '健身', '休闲活动；', '；城市绿化管理；商务秘书服务。（除依法须', '经批准的项目外，凭营业执照依法自主开展经营活动）', '（一般经营项目自主经营，许可经营项目凭相关许可证', '登记机关', '或者批准文件经营）（依法须经批准的项目，经相关部', '门批准后方可开展经营活动。）', '01433969', '2022', '年', '0', '月', '2周家企业信用信息公示系统网址：', '第1页/共1页', '国家市场监督管理总局监制']
    location = [[282, 723], [438, 719], [541, 721], [650, 721], [797, 721], [1040, 720], [1382, 719], [451, 748], [709, 747], [1191, 755], [654, 773], [655, 801], [655, 827], [561, 854], [820, 853], [506, 879], [670, 879], [793, 879], [887, 880], [441, 907], [700, 906], [649, 933], [660, 959], [1192, 955], [655, 986], [540, 1013], [293, 1056], [1325, 1046], [1376, 1056], [1442, 1047], [1461, 1054], [216, 1181], [889, 1188], [1532, 1178]]
    centers, labels = kmeans(np.asarray(location), n_clusters=2, show=True)
    x_list = centers[:, 0].tolist()
    # 获取最左侧的簇
    index = x_list.index(min(x_list))
    labels = labels.tolist()
    result = []
    others = []
    print("".join(txt))
    print("-----------")
    for i in range(0, len(labels)):
        if labels[i] == index:
            result.append(txt[i])
        else:
            others.append(txt[i])
    print("".join(result))
    print("-----------")
    print("".join(others))
    print(labels)



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2022/2/28 
# @Function:
import time
from pathlib import Path
import warnings

import jieba
import pandas as pd
import numpy as np
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.model_selection import GridSearchCV
from sklearn.preprocessing import OneHotEncoder

from zhousflib.ml.model_base import BaseModel
from zhousflib.ml.feature_vector import FeatureVector, TypeFeatureVector


class GBDTClassifier(BaseModel):

    def __init__(self, model_file, vector_pickle, vector_type=TypeFeatureVector.TYPE_TF_IDF,
                 importance_files=None, is_slim_vector=False):
        """

        :param model_file: 模型文件
        :param vector_pickle: 向量文件
        :param vector_type: 特征向量表示器
        :param importance_files: 不重要特征文件
        :param is_slim_vector: True时使用importance_files
        """
        self.model_file = model_file
        if vector_pickle is None:
            warnings.warn("vector_pickle is None.")
        if model_file is None:
            warnings.warn("model_file is None.")
        self.vector = FeatureVector(default_pkl=vector_pickle, vector_type=vector_type)
        model = self.load_model(model_file) if os.path.exists(model_file) else None
        self.params = {}
        self.words_not_importance = self.read_words_not_importance(importance_files) if is_slim_vector else []
        super().__init__(model, model_file)

    def build_feature(self, excel_files: list, one_hot_class=False):
        print("构建数据特征...")
        dfs = []
        for excel_file in excel_files:
            df = pd.read_excel(excel_file, header=0)
            dfs.append(df)
        data_frame = pd.concat(dfs)
        data_frame = data_frame.fillna("")
        class_attr_name = "type"
        if one_hot_class:
            # 将类别进行OneHotEncoder编码
            class_value_list = np.arange(3).tolist()
            hot_en = OneHotEncoder(categories=[class_value_list])
            y = hot_en.fit_transform(data_frame[class_attr_name].values.reshape(-1, 1)).toarray()
        else:
            y = data_frame[class_attr_name]
        features = []
        data = []
        for index, row in data_frame.iterrows():
            row_list = list(row)
            name = row_list[0]
            label_gt = row_list[-2]
            full_str = row_list[-1]
            full_str = full_str.replace(" ", "")
            full_str = jieba.lcut(full_str)
            # 删除指定特征
            for word in full_str:
                if word in self.words_not_importance:
                    full_str.remove(word)
            data.append(" ".join(full_str))
            vec_tf = self.vector.transform([" ".join(full_str)])
            features.append((name, label_gt, vec_tf))
        # 数据向量化
        vec_tf = self.vector.transform(data)
        x = vec_tf.toarray()
        return features, x, y, self.vector.vocabulary, data

    def get_top(self, predict_prob_result, top):
        prob_dict = {}
        predict_result = predict_prob_result[0]
        for i in range(0, len(predict_result)):
            prob_dict[i] = predict_result[i]
        result = sorted(prob_dict.items(), key=lambda p: p[1], reverse=True)
        top = len(result) if top > len(result) else top
        return result[:top]

    def __search_n_estimators(self, x_train, y_train):
        print("search n_estimators...")
        param = {'n_estimators': range(20, 81, 10)}
        estimator = GradientBoostingClassifier(learning_rate=0.1, min_samples_split=300,
                                               min_samples_leaf=20, max_depth=8,
                                               max_features='sqrt', subsample=0.8,
                                               random_state=10)
        search = GridSearchCV(estimator=estimator, param_grid=param, cv=5)
        search.fit(x_train, y_train)
        print(search.best_params_)
        print(search.best_score_)
        self.params.update(search.best_params_)
        """
        {'n_estimators': 80}
        0.9808457711442786
        """

    def __search_max_depth(self, x_train, y_train):
        print("search max_depth...")
        param = {'max_depth': range(3, 14, 2), 'min_samples_split': range(100, 801, 200)}
        estimator = GradientBoostingClassifier(learning_rate=0.1,
                                               n_estimators=self.params.get("n_estimators"),
                                               min_samples_leaf=20, max_features='sqrt', subsample=0.8, random_state=10)
        search = GridSearchCV(estimator=estimator, param_grid=param, cv=5)
        search.fit(x_train, y_train)
        print(search.best_params_)
        print(search.best_score_)
        self.params.update(search.best_params_)
        """
        {'max_depth': 13, 'min_samples_split': 500}
        0.985323383084577
        """

    def __search_min_sample(self, x_train, y_train):
        print("search min_sample...")
        param = {'min_samples_split': range(800, 1900, 200), 'min_samples_leaf': range(60, 101, 10)}
        estimator = GradientBoostingClassifier(learning_rate=0.1,
                                               n_estimators=self.params.get("n_estimators"),
                                               max_depth=self.params.get("max_depth"),
                                               min_samples_leaf=20, max_features='sqrt', subsample=0.8, random_state=10)
        search = GridSearchCV(estimator=estimator, param_grid=param, cv=5)
        search.fit(x_train, y_train)
        print(search.best_params_)
        print(search.best_score_)
        self.params.update(search.best_params_)
        """
        {'min_samples_leaf': 60, 'min_samples_split': 1000}
        0.9713930348258707
        """

    def __search_max_features(self, x_train, y_train):
        print("search max_features...")
        param = {'max_features': range(7, 20, 2)}
        estimator = GradientBoostingClassifier(learning_rate=0.1,
                                               n_estimators=self.params.get("n_estimators"),
                                               max_depth=self.params.get("max_depth"),
                                               min_samples_split=self.params.get("min_samples_split"),
                                               min_samples_leaf=self.params.get("min_samples_leaf"),
                                               max_features='sqrt', subsample=0.8, random_state=10)
        search = GridSearchCV(estimator=estimator, param_grid=param, cv=5)
        search.fit(x_train, y_train)
        print(search.best_params_)
        print(search.best_score_)
        self.params.update(search.best_params_)
        """
        {'max_features': 19}
        0.8213930348258707
        """

    def __search_subsample(self, x_train, y_train):
        print("search subsample...")
        param = {'subsample': [0.6, 0.7, 0.75, 0.8, 0.85, 0.9]}
        estimator = GradientBoostingClassifier(learning_rate=0.1,
                                               n_estimators=self.params.get("n_estimators"),
                                               max_depth=self.params.get("max_depth"),
                                               min_samples_split=self.params.get("min_samples_split"),
                                               min_samples_leaf=self.params.get("min_samples_leaf"),
                                               max_features=self.params.get("max_features"),
                                               subsample=0.8, random_state=10)
        search = GridSearchCV(estimator=estimator, param_grid=param, cv=5)
        search.fit(x_train, y_train)
        print(search.best_params_)
        print(search.best_score_)
        self.params.update(search.best_params_)
        """
        {'subsample': 0.9}
        0.8281094527363184
        """

    def search_best_params(self, excel_file):
        x_train, x_test, y_train, y_test = self.load_data(excel_file, one_hot_class=False, shuffle=True, test_size=0.2)
        self.__search_n_estimators(x_train, y_train)
        self.__search_max_depth(x_train, y_train)
        self.__search_min_sample(x_train, y_train)
        self.__search_max_features(x_train, y_train)
        self.__search_subsample(x_train, y_train)
        print("search_best_params...")
        print(self.params)

    def predict(self, string, top=1):
        """
        预测
        :param string:
        :param top:
        :return: label_pre, confidence
        """
        data = []
        full_str = string.replace(" ", "")
        full_str = jieba.lcut(full_str)
        # 删除指定特征
        for word in full_str:
            if word in self.words_not_importance:
                full_str.remove(word)
        data.append(" ".join(full_str))
        prob = self.model.predict_proba(self.vector.transform(data))
        res = self.get_top(prob, top=top)[0]
        return res

    def show_feature_importance(self):
        print(self.model.feature_importances_)
        print(len(self.model.feature_importances_))
        vec_word = []
        # 将特征字典排序
        word_vec = sorted(self.vector.vocabulary.items(), key=lambda x: x[1], reverse=False)
        for word in word_vec:
            w, index = word
            vec_word.append(w)
        word_import = {}
        for i in range(0, len(self.model.feature_importances_)):
            word_import[vec_word[i]] = self.model.feature_importances_[i]
        # 根据特征重要性倒叙
        word_import = sorted(word_import.items(), key=lambda x: x[1], reverse=True)
        write_data = []
        for j in word_import:
            write_data.append("{0} {1}".format(j[0], j[1]))
        importance_file = Path.cwd().joinpath("importance_gbdt.txt")
        importance_file.write_text("\n".join(write_data))

    def get_model(self):
        return GradientBoostingClassifier(learning_rate=0.1, n_estimators=80, max_depth=19,
                                          min_samples_leaf=20, max_features='sqrt', subsample=0.8, random_state=200)


if __name__ == "__main__":
    # --------- 1. 搜索最佳参数
    # classifier.search_best_params(excel_file=excel_file_)
    # --------- 2. 训练
    # classifier.train(excel_files=excel_files_, one_hot_class=False, shuffle=True, test_size=0.2)
    # --------- 3. 预测
    # print(classifier.predict(""))
    # --------- 4. 测试集评估
    # classifier.val_test(excel_file="", classes=labels)
    # --------- 5. 展示特征的重要程度
    # classifier.show_feature_importance()
    # --------- 6. 卡方校验
    # classifier.chi_square_check(excel_file=excel_file_, one_hot_class=False, k=1000)
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/19 
# @Function:
import torch

from backend import Backend


class BackendTorchScript(Backend):
    def __init__(self, *args, **kwargs):
        self.model_dir = kwargs.get("model_dir")
        self.device_id = kwargs.get("device_id")
        self.model = None
        self.backend = "torch_script"
        self.device = torch.device('cuda:{0}'.format(self.device_id) if self.device_id > -1 else 'cpu')
        super().__init__(*args)

    def build(self, **kwargs):
        module = kwargs.get("module")
        example_inputs = kwargs.get("example_inputs")

        self.pop(kwargs, "from_platform")
        self.pop(kwargs, "module")
        self.pop(kwargs, "dynamic_axes")
        self.pop(kwargs, "opset_version")
        self.pop(kwargs, "example_inputs")
        self.pop(kwargs, "shape")

        self.pop(kwargs, "model_name")
        self.pop(kwargs, "model_version")
        self.pop(kwargs, "http_inputs")
        self.pop(kwargs, "http_outputs")
        self.pop(kwargs, "concurrency")

        target_files = self.get_file_by_suffix(model_dir=self.model_dir, suffix=".pt")
        if len(target_files) == 0:
            assert example_inputs, "导出torchScript时，example_inputs不能为空."
            assert module, "导出torchScript时，module不能为空."
            print("torchScript文件不存在，正在导出torchScript...")
            from zhousflib.ann.torch import torch_to_script
            torch_to_script.convert_script_model(module=module, model_dir=self.model_dir, export_dir=self.model_dir,
                                                 device_id=self.device_id, example_inputs=example_inputs)
            target_files_ = self.get_file_by_suffix(model_dir=self.model_dir, suffix=".pt")
            if len(target_files_) > 0:
                print("导出torchScript文件成功：{0}".format(target_files_[0]))
                target_files.append(target_files_[0])
        else:
            print("torchScript文件存在：{0}.".format(target_files[0]))

        if kwargs.get("autoload", True):
            from zhousflib.ann.torch import torch_to_script
            self.model, _, _ = torch_to_script.load_script_model(self.model_dir, device_id=self.device_id)
            print("加载torchscript成功：{0}.".format(target_files[0]))
        return target_files[0]

    def inference(self, inputs_list: list):
        inputs = []
        for i, input_ in enumerate(inputs_list):
            inputs.append(inputs_list[i].squeeze(1).to(self.device))
        return self.model(inputs)


# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/19 
# @Function:
import abc
from pathlib import Path


class Backend(metaclass=abc.ABCMeta):

    def __init__(self, *args):
        pass

    @abc.abstractmethod
    def build(self, **kwargs):
        pass

    @abc.abstractmethod
    def inference(self, inputs_list: list):
        pass

    @staticmethod
    def to_numpy(tensor):
        return tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()

    @staticmethod
    def get_file_by_suffix(model_dir: Path, suffix: str):
        return [file for file in model_dir.glob("*{0}".format(suffix))]

    @staticmethod
    def pop(kwargs: dict, key: str):
        if len(kwargs) == 0 or len(key) == 0:
            return
        if key in kwargs:
            kwargs.pop(key)


# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/11/1 
# @Function:
import shutil
from pathlib import Path


def overwrite_folder(from_dir: Path, to_dir: Path):
    """
    覆盖目录，覆盖已有文件
    :param from_dir: 目录
    :param to_dir: 被覆盖目录
    :return:
    """
    for file in from_dir.rglob("*.*"):
        shutil.copy(file, to_dir.joinpath(file.parent.name))
        print(file)
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2022/9/21
# @Function: 特征向量表示器
import jieba
import pickle
import os
import pandas as pd
from enum import Enum
from pathlib import Path
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.feature_extraction.text import CountVectorizer


class TypeFeatureVector(Enum):
    """
    特征向量器
    """
    TYPE_TF_IDF = 1  # TfidfVectorizer
    TYPE_COUNT_VECTOR = 2  # CountVectorizer

    @staticmethod
    def get_type(type_name):
        if type_name == TypeFeatureVector.TYPE_TF_IDF.name:
            return TypeFeatureVector.TYPE_TF_IDF
        elif type_name == TypeFeatureVector.TYPE_COUNT_VECTOR.name:
            return TypeFeatureVector.TYPE_COUNT_VECTOR
        return None


class FeatureVector(object):
    """
    特征向量表示器
    """

    def __init__(self, default_pkl, vector_type=TypeFeatureVector.TYPE_TF_IDF, force_update=False):
        """

        :param default_pkl: 特征向量文件
        :param vector_type: 特征向量类型
        :param force_update: 强制更新模型文件
        """
        self.default_pkl = default_pkl
        self.vector = self.load_model(default_pkl) if not force_update else None
        if self.vector is None:
            if vector_type == TypeFeatureVector.TYPE_TF_IDF:
                self.vector = TfidfVectorizer(lowercase=True)
            elif vector_type == TypeFeatureVector.TYPE_COUNT_VECTOR:
                self.vector = CountVectorizer(lowercase=True)
        self.vector_type = vector_type
        self.vocabulary = self.vector.vocabulary_ if hasattr(self.vector, "vocabulary_") else None

    @staticmethod
    def save_model(model, model_file="vector.pickle"):
        # 保存模型
        pickle.dump(model, open(model_file, "wb"))

    @staticmethod
    def load_model(model_file):
        # 加载模型
        if model_file is None or not os.path.exists(model_file):
            return None
        return pickle.load(open(model_file, "rb"))

    def build_save_vector(self, text_list):
        """
        构建字典表-全量
        :param text_list:
        :return:
        """
        model = self.vector.fit(text_list)
        # 保存向量器
        self.save_model(model, self.default_pkl)
        return self.vector

    def transform(self, text_list):
        """
        向量化
        :param text_list:
        :return:
        """
        vec_form = self.vector.transform(text_list)
        return vec_form


def build_dict(excel_files, importance_files=None, save_vector_file=None, vector_type=TypeFeatureVector.TYPE_TF_IDF):
    """
    构建数据集字典表
    :param excel_files: 全量数据集
    :param importance_files: 不重要特征集合
    :param save_vector_file: 保存特征向量文件
    :param vector_type: 特征向量类型
    :return:
    """
    # 构建不重要特征集合
    words_not_importance = []
    if importance_files is not None:
        for importance_file in importance_files:
            importance_file = Path(importance_file)
            txt = importance_file.read_text()
            for line in txt.split("\n"):
                word, importance = line.split(" ")
                # importance等于0
                if float(importance) == 0:
                    words_not_importance.append(word)
    # 构建字典表
    data_frames = []
    for excel_file in excel_files:
        data_frames.append(pd.read_excel(excel_file, header=0))
    data_frame = pd.concat(data_frames)
    data_frame = data_frame.fillna("")
    data = []
    for index, row in data_frame.iterrows():
        row_list = list(row)
        full_str = row_list[-1]
        full_str = full_str.replace(" ", "")
        full_str = jieba.lcut(full_str)
        # 删除指定特征
        for word in full_str:
            if word in words_not_importance:
                full_str.remove(word)
        data.append(" ".join(full_str))
    # 初始化特征向量表示器
    save_vector_file = "vector_tfidf.pickle" if save_vector_file is None else save_vector_file
    vec_f = FeatureVector(default_pkl=save_vector_file,
                          vector_type=vector_type,
                          force_update=True)
    vec_f.build_save_vector(data)
    print(len(vec_f.vector.vocabulary_))
    print("done.")


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/18 
# @Function:
from pathlib import Path


class FastInfer(object):

    def __init__(self, model_dir: Path, device_id=-1):
        """
        :param model_dir:
        :param device_id: cpu上运行：-1 | gpu上运行：0 or 1 or 2...
        """
        self.model_dir = model_dir
        self.device_id = device_id
        self.backend = None

    def use_http_backend(self, url: str, model_name: str, model_version: str, http_inputs: list, http_outputs: list, **kwargs):
        """
        设置推理后端：http
        :param url:
        :param model_name:
        :param model_version:
        :param http_inputs:
        :param http_outputs:
        :param kwargs:
        :return:
        example:
        .use_http_backend(url="127.0.0.1:5005",
                          model_name="cosnet_onnx",
                          model_version="1",
                          http_inputs=[("input_ids", "INT64"), ("token_type_ids", "INT64"), ("attention_mask", "INT64")],
                          http_outputs=["output"])
        """
        from backend_http import BackendHttp
        self.backend = BackendHttp(model_dir=self.model_dir, device_id=self.device_id)
        return self.backend.build(url=url, model_name=model_name, model_version=model_version, http_inputs=http_inputs, http_outputs=http_outputs, **kwargs)

    def use_original_backend(self, module=None, **kwargs):
        """
        设置推理后端：原始模型-动态图
        :param module:
        :param kwargs:
        :return:
        example:
        .use_original_backend(module=torch.nn.Module())
        """
        from backend_original import BackendOriginal
        self.backend = BackendOriginal(model_dir=self.model_dir, device_id=self.device_id)
        return self.backend.build(module=module, **kwargs)

    def use_onnx_backend(self, from_platform: str = None, module=None, dynamic_axes: dict = None, opset_version=10, example_inputs=None, **kwargs):
        """
        设置推理后端：onnxruntime
        :param from_platform:
        :param module:
        :param dynamic_axes:
        :param opset_version:
        :param example_inputs:
        :param kwargs:
        :return:
        example:
        .use_onnx_backend(from_platform="torch",
                          module=torch.nn.Module(),
                          example_inputs=example_inputs,
                          dynamic_axes={'input_ids': {0: 'batch_size'},
                                        'token_type_ids': {0: 'batch_size'},
                                        'attention_mask': {0: 'batch_size'},
                                        'output': {0: 'batch_size'}})
        """
        from backend_onnx import BackendONNX
        self.backend = BackendONNX(model_dir=self.model_dir, device_id=self.device_id)
        return self.backend.build(from_platform=from_platform, module=module, dynamic_axes=dynamic_axes, opset_version=opset_version, example_inputs=example_inputs, **kwargs)

    def use_tensorrt_backend(self, from_platform: str = None, module=None, dynamic_axes: dict = None, opset_version=10,
                             example_inputs=None, shape: dict = None, **kwargs):
        """
        设置推理后端：tensorrt
        :param from_platform:
        :param module:
        :param dynamic_axes:
        :param opset_version:
        :param example_inputs:
        :param shape:
        :param kwargs:
        :return:
        example:
        若onnx文件存在时则只需传参数shape
        .use_tensorrt_backend(shape={"input_ids": [(10, 128), (10, 128), (10, 128)],
                                     "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                                     "attention_mask": [(10, 128), (10, 128), (10, 128)]})
        若onnx文件不存在时
        .use_tensorrt_backend(from_platform="torch",
                              module=torch.nn.Module(),
                              example_inputs=example_inputs,
                              dynamic_axes={'input_ids': {0: 'batch_size'},
                                          'token_type_ids': {0: 'batch_size'},
                                          'attention_mask': {0: 'batch_size'},
                                          'output': {0: 'batch_size'}},
                              shape={"input_ids": [(10, 128), (10, 128), (10, 128)],
                                     "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                                     "attention_mask": [(10, 128), (10, 128), (10, 128)]})
        """
        from backend_tensorrt import BackendTensorRT
        self.backend = BackendTensorRT(model_dir=self.model_dir, device_id=self.device_id)
        return self.backend.build(from_platform=from_platform, module=module, dynamic_axes=dynamic_axes,
                                  opset_version=opset_version, example_inputs=example_inputs, shape=shape, **kwargs)

    def use_torch_script_backend(self, module=None, example_inputs=None, **kwargs):
        """
        设置推理后端：torch_script
        :param module:
        :param example_inputs:
        :param kwargs:
        :return:
        example:
        .use_torch_script_backend(module=torch.nn.Module(), example_inputs=example_inputs)
        """
        from backend_torch_script import BackendTorchScript
        self.backend = BackendTorchScript(model_dir=self.model_dir, device_id=self.device_id)
        return self.backend.build(module=module, example_inputs=example_inputs, **kwargs)

    def infer(self, inputs_list: list):
        assert self.backend is not None, "请设置backend，例如.use_onnx_backend()"
        return self.backend.inference(inputs_list=inputs_list)

    @staticmethod
    def demo():
        import torch
        from zhousflib.ann.torch import torch_to_onnx
        fast_infer = FastInfer(model_dir=Path(r"F:\torch\test"), device_id=0)
        fast_infer.use_tensorrt_backend(from_platform="torch",
                                        module=torch.nn.Module(),
                                        example_inputs=(torch_to_onnx.example_inputs_demo(device_id=0),),
                                        dynamic_axes={'input_ids': {0: 'batch_size'},
                                                      'token_type_ids': {0: 'batch_size'},
                                                      'attention_mask': {0: 'batch_size'},
                                                      'output': {0: 'batch_size'}},
                                        shape={"input_ids": [(10, 128), (10, 128), (10, 128)],
                                               "token_type_ids": [(10, 128), (10, 128), (10, 128)],
                                               "attention_mask": [(10, 128), (10, 128), (10, 128)]})
        return fast_infer.infer(torch_to_onnx.example_inputs_demo()).tolist()




# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/19 
# @Function:
from backend import Backend


class BackendONNX(Backend):
    def __init__(self, *args, **kwargs):
        self.model_dir = kwargs.get("model_dir")
        self.device_id = kwargs.get("device_id")
        self.model = None
        self.backend = "onnx_runtime"
        super().__init__(*args)

    def build(self, **kwargs):
        from_platform = kwargs.get("from_platform")
        module = kwargs.get("module")
        dynamic_axes = kwargs.get("dynamic_axes")
        opset_version = kwargs.get("opset_version")
        example_inputs = kwargs.get("example_inputs")

        self.pop(kwargs, "from_platform")
        self.pop(kwargs, "module")
        self.pop(kwargs, "dynamic_axes")
        self.pop(kwargs, "opset_version")
        self.pop(kwargs, "example_inputs")
        self.pop(kwargs, "shape")

        self.pop(kwargs, "model_name")
        self.pop(kwargs, "model_version")
        self.pop(kwargs, "http_inputs")
        self.pop(kwargs, "http_outputs")
        self.pop(kwargs, "concurrency")

        target_files = self.get_file_by_suffix(model_dir=self.model_dir, suffix=".onnx")
        if len(target_files) == 0:
            assert from_platform in ["torch"], "暂不支持{0}平台".format(from_platform)
            print("onnx文件不存在，正在导出onnx...")
            if from_platform == "torch":
                assert example_inputs, "导出onnx时，example_inputs不能为空."
                assert module, "导出onnx时，module不能为空."
                bin_files = self.get_file_by_suffix(model_dir=self.model_dir, suffix=".bin")
                assert len(bin_files) > 0, '未找到.bin权重文件，请检查模型目录：{0}.'.format(self.model_dir)
                import torch
                from zhousflib.ann import get_device
                state_dict = torch.load(bin_files[0], map_location=get_device(device_id=self.device_id))
                module.load_state_dict(state_dict)
                module.eval()
                output_names = []
                input_names = []
                for name in dynamic_axes.keys():
                    if str(name).startswith("output"):
                        output_names.append(name)
                    else:
                        input_names.append(name)
                assert len(output_names) > 0, "在dynamic_axes中未找到output：{0}.".format(dynamic_axes)
                save_file = self.model_dir.joinpath("model.onnx")
                torch.onnx.export(model=module, args=example_inputs, f=str(save_file),
                                  opset_version=opset_version, input_names=input_names, output_names=output_names,
                                  dynamic_axes=dynamic_axes, **kwargs)
                print("导出onnx文件成功：{0}".format(save_file))
                target_files.append(save_file)
        else:
            print("onnx文件存在：{0}.".format(target_files[0]))

        if kwargs.get("autoload", True):
            from zhousflib.ann.torch import torch_to_onnx
            self.model, _, _ = torch_to_onnx.load_onnx(self.model_dir, device_id=self.device_id, autoload_weights=False, autoload_tokenizer=False)
        print("加载onnx成功：{0}.".format(target_files[0]))
        return target_files[0]

    def inference(self, inputs_list: list):
        feed = {}
        for i, input_ in enumerate(inputs_list):
            feed[self.model.get_inputs()[i].name] = self.to_numpy(inputs_list[i])
        result = self.model.run(None, feed)
        return result[0]


# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/7 
# @Function: 参考 https://github.com/NVIDIA/TensorRT/blob/master/samples/python/common.py

import os
os.environ["CUDA_MODULE_LOADING"] = "LAZY"
from pathlib import Path

import numpy as np
import pycuda.driver as cuda
import pycuda.autoinit
import tensorrt as trt

from zhousflib.ann import check_device_id

"""
ERROR: INVALID_CONFIG: The engine plan file is generated on an incompatible device, 
expecting compute 6.1 got compute 7.5, please rebuild. 
解决方法：在线上重新进行TensorRT模型的转换(本地机器上的GPU与线上机器的GPU版本不一致，导致的加速版本不一致)
"""

try:
    # Sometimes python does not understand FileNotFoundError
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


# Simple helper data class that's a little nicer to use than a 2-tuple.
class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


class RTInfer(object):

    def __init__(self, trt_file_path: Path, device_id: int = -1, use_stack=False):
        """

        :param trt_file_path: trt文件路径
        :param device_id: cpu上运行：-1 | gpu上运行：0 or 1 or 2...
        :param use_stack: 是否使用压栈策略，牺牲并发性能
        """
        check_device_id(device_id)
        self.cuda_ctx = cuda.Device(device_id).make_context() if use_stack else None
        with trt_file_path.open("rb") as f:
            runtime = trt.Runtime(trt.Logger(trt.Logger.WARNING))
            self.engine = runtime.deserialize_cuda_engine(f.read())
            self.context = self.engine.create_execution_context()

    @staticmethod
    def allocate_buffers(engine):
        inputs = []
        outputs = []
        bindings = []
        stream = cuda.Stream()
        for binding in engine:
            size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
            dtype = trt.nptype(engine.get_binding_dtype(binding))
            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            # Append the device buffer to device bindings.
            bindings.append(int(device_mem))
            # Append to the appropriate list.
            if engine.binding_is_input(binding):
                inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                outputs.append(HostDeviceMem(host_mem, device_mem))
        return inputs, outputs, bindings, stream

    # This function is generalized for multiple inputs/outputs.
    # inputs and outputs are expected to be lists of HostDeviceMem objects.
    def do_inference(self, context, bindings, inputs, outputs, stream, batch_size=1):
        # Transfer input data to the GPU.
        [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
        if self.cuda_ctx:
            self.cuda_ctx.push()
        # Run inference.
        context.execute_async(batch_size=batch_size, bindings=bindings, stream_handle=stream.handle)
        if self.cuda_ctx:
            self.cuda_ctx.pop()
        # Transfer predictions back from the GPU.
        [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
        # Synchronize the stream
        stream.synchronize()
        # Return only the host outputs.
        return [out.host for out in outputs]

    # This function is generalized for multiple inputs/outputs for full dimension networks.
    # inputs and outputs are expected to be lists of HostDeviceMem objects.
    def do_inference_v2(self, context, bindings, inputs, outputs, stream):
        # Transfer input data to the GPU.
        [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
        if self.cuda_ctx:
            self.cuda_ctx.push()
        # Run inference.
        context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
        if self.cuda_ctx:
            self.cuda_ctx.pop()
        # Transfer predictions back from the GPU.
        [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
        # Synchronize the stream
        stream.synchronize()
        # Return only the host outputs.
        return [out.host for out in outputs]

    def infer(self, input_arr: np.asarray, batch_size: int = None):
        """
        推理
        :param input_arr: input_arr=np.asarray([to_numpy(args[0].int()), to_numpy(args[1].int()), to_numpy(args[2].int())])
        :param batch_size:
        :return:
        """
        self.context.active_optimization_profile = 0
        for i, item in enumerate(input_arr):
            self.context.set_binding_shape(i, item.shape)
        inputs, outputs, bindings, stream = self.allocate_buffers(self.engine)
        for i, item in enumerate(input_arr):
            inputs[i].host = item
        if batch_size:
            return self.do_inference(self.context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream, batch_size=batch_size)
        else:
            return self.do_inference_v2(self.context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream)


if __name__ == "__main__":
    from zhousflib.ann import to_numpy
    from zhousflib.ann.torch.torch_to_onnx import example_inputs_demo
    args = example_inputs_demo(input_size=1)
    batch = np.asarray([to_numpy(args[0].int()), to_numpy(args[1].int()), to_numpy(args[2].int())])
    rt_engine = RTInfer(trt_file_path=Path(r"F:\torch\onnx\model_32.trt"), device_id=0, use_stack=True)
    data = rt_engine.infer(input_arr=batch)
    print(data)



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/11/27 
# @Function:
import shutil
from pathlib import Path

import torch
import onnxruntime

from zhousflib.ann import to_numpy
from zhousflib.ann import transformers
from zhousflib.ann.torch import check_cuda, get_device, check_device_id


def load_onnx(model_dir: Path, device_id: int = -1, autoload_weights=True, autoload_tokenizer=True):
    """
    加载onnx模型
    :param model_dir: 模型目录
    :param device_id: cpu上运行：-1 | gpu上运行：0 or 1 or 2...
    :param autoload_weights: 加载权重
    :param autoload_tokenizer: 加载tokenizer
    :return:
    ort_session, _, _ = load_onnx(model_dir=Path(r"F:\torch\onnx"))
    ort_input = ort_session.get_inputs()
    args = example_inputs_demo()
    ort_inputs = {ort_input[0].name: to_numpy(args[0]),
                  ort_input[1].name: to_numpy(args[1]),
                  ort_input[2].name: to_numpy(args[2])}
    ort_outs = ort_session.run(None, ort_inputs)
    print(ort_outs[0])
    """
    onnx_file = None
    tokenizer = None
    state_dict = None
    for file in model_dir.glob("*.onnx"):
        onnx_file = file
        break
    assert onnx_file is not None, ".onnx file not found in directory{0}".format(model_dir)
    if device_id == -1:
        onnx_session = onnxruntime.InferenceSession(str(onnx_file))
    else:
        onnx_session = onnxruntime.InferenceSession(str(onnx_file), providers=['CUDAExecutionProvider'],
                                                    provider_options=[{'device_id': device_id}])
    if autoload_weights:
        state_dict = load_weights(model_dir=model_dir, device_id=device_id)
    if autoload_tokenizer:
        tokenizer = transformers.load_tokenizer(model_dir=model_dir)
    return onnx_session, tokenizer, state_dict


def load_weights(model_dir: Path, device_id: int = -1):
    """
    加载模型权重
    :param model_dir: 模型目录
    :param device_id: cpu上运行：-1 | gpu上运行：0 or 1 or 2...
    :return:
    """
    bin_file = None
    for file in model_dir.glob("*.bin"):
        bin_file = file
    if bin_file:
        state_dict = torch.load(bin_file, map_location=get_device(device_id))
        return state_dict
    else:
        return None


def convert_onnx(model_dir: Path, export_dir: Path, device_id: int = -1, example_inputs=None, module: torch.nn.Module = None, **kwargs):
    """
    导出onnx
    :param model_dir: 模型目录
    :param export_dir: 导出目录
    :param device_id: 绑定硬件, cpu上运行：-1 | gpu上运行：0 or 1 or 2...
    :param example_inputs: 输入示例
    :param module: 神经网络
    :param kwargs: 自定义参数
    :return:
    """
    bin_file = None
    if not export_dir.exists():
        export_dir.mkdir(parents=True)
    if module is None:
        config = transformers.load_config(model_dir=model_dir)
        model = transformers.load_model(model_dir=model_dir, config=config)
    else:
        model = module
    # 权重文件，这个是给预测的后处理模块初始化权重文件做准备
    for file in model_dir.glob("*.bin"):
        bin_file = file
        if not export_dir.joinpath(bin_file.name).exists():
            shutil.copy(bin_file, export_dir)
            break
    if module is not None:
        assert bin_file, '.bin file is not exists, please check {0}.'.format(model_dir)
        state_dict = torch.load(bin_file, map_location=get_device(device_id))
        model.load_state_dict(state_dict)
    model.eval()
    torch.onnx.export(model, example_inputs, str(export_dir.joinpath("model.onnx")), **kwargs)
    # tokenizer文件，这个是给预测的input data做准备
    tokenizer = transformers.load_tokenizer(model_dir=model_dir)
    if tokenizer is not None:
        tokenizer.save_pretrained(export_dir)
    print("done.")


def example_inputs_demo(device_id: int = -1, input_size=10, batch_size=128):
    """
    输入示例
    :param device_id: cpu上运行：-1 | gpu上运行：0 or 1 or 2...
    :param input_size:
    :param batch_size:
    :return:
    """
    check_device_id(device_id)
    ids = torch.LongTensor(input_size, batch_size).zero_()
    seq_len = torch.LongTensor(input_size, batch_size).zero_()
    mask = torch.LongTensor(input_size, batch_size).zero_()
    if device_id == -1:
        return [ids, seq_len, mask]
    else:
        check_cuda()
        return [ids.cuda(device_id), seq_len.cuda(device_id), mask.cuda(device_id)]


def convert_bert_demo():
    """
    转换示例：以bert转onnx为例
    :return:
    """
    """
    通用导出示例
    """
    # convert_onnx(module=torch.nn.Module(),
    #              model_dir=Path(r"F:\torch\train_model"),
    #              export_dir=Path(r"F:\torch\onnx2"),
    #              device="cpu", example_inputs=(example_inputs_demo(device_id=-1), ),
    #              verbose=True,
    #              export_params=True,
    #              opset_version=10,
    #              input_names=['input_ids', 'token_type_ids', 'attention_mask'],
    #              output_names=['output'],
    #              dynamic_axes={'input_ids': {0: 'batch_size'},
    #                            'token_type_ids': {0: 'batch_size'},
    #                            'attention_mask': {0: 'batch_size'},
    #                            'output': {0: 'batch_size'}})
    """
    自定义导出示例（以bert导出为例）
    """
    args = example_inputs_demo(device_id=-1)
    args = args[0], args[1], args[2],
    convert_onnx(model_dir=Path(r"F:\torch\train_model"),
                 export_dir=Path(r"F:\torch\onnx"),
                 example_inputs=args,
                 verbose=True,
                 export_params=True,
                 opset_version=10,
                 input_names=['input_ids', 'token_type_ids', 'attention_mask'],
                 output_names=['output'],
                 dynamic_axes={'input_ids': {0: 'batch_size'},
                               'token_type_ids': {0: 'batch_size'},
                               'attention_mask': {0: 'batch_size'},
                               'output': {0: 'batch_size'}})


if __name__ == "__main__":
    """
    onnx转换示例
    """
    # convert_bert_demo()
    """
    onnx预测示例
    """
    ort_session, _, _ = load_onnx(model_dir=Path(r"F:\torch\onnx"), device_id=0)
    ort_input = ort_session.get_inputs()
    args = example_inputs_demo(input_size=32)
    feed = {}
    for i, input_ in enumerate(args):
        feed[ort_input.get_inputs()[i].name] = to_numpy(args[i])
    options = onnxruntime.RunOptions()
    ort_outs = ort_session.run(None, feed, run_options=options)
    print(ort_outs)
    # pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/12/14 
# @Function:
import time

from gevent import monkey
monkey.patch_all()
import numpy as np
import gevent.ssl
import tritonclient.http as httpclient
from tritonclient.utils import InferenceServerException


"""
############## 【安装 triton client】 ##############
文档：https://github.com/triton-inference-server/client
参考：https://github.com/triton-inference-server/client/blob/main/src/python/examples/simple_http_async_infer_client.py
pip install tritonclient
pip install geventhttpclient
or
pip install tritonclient[all]
sudo apt update
sudo apt install libb64-dev
"""


class ClientHttp(object):
    def __init__(self, url="localhost:8080", concurrency=1, connection_timeout=60.0, network_timeout=60.0, verbose=False,
                 ssl=False, key_file: str = None, cert_file: str = None, ca_certs: str = None, insecure=False):
        """
        创建客户端
        :param url: Inference server URL. Default is localhost:8000.
        :param concurrency: 并发数量
        :param connection_timeout: 连接超时
        :param network_timeout: 网络超时
        :param verbose: Enable verbose output
        :param ssl: Enable encrypted link to the server using HTTPS
        :param key_file: File holding client private key. Default is None.
        :param cert_file: File holding client certificate. Default is None.
        :param ca_certs: File holding ca certificate. Default is None.
        :param insecure: Use no peer verification in SSL communications. Use with caution. Default is False.
        :return:
        """
        self.url = url
        self.concurrency = concurrency
        self.connection_timeout = connection_timeout
        self.network_timeout = network_timeout
        self.verbose = verbose
        self.ssl = ssl
        self.key_file = key_file
        self.cert_file = cert_file
        self.ca_certs = ca_certs
        self.insecure = insecure
        self.client = self._create()

    def _create(self):
        try:
            if self.ssl:
                ssl_options = {}
                if self.key_file is not None:
                    ssl_options["keyfile"] = self.key_file
                if self.cert_file is not None:
                    ssl_options["certfile"] = self.cert_file
                if self.ca_certs is not None:
                    ssl_options["ca_certs"] = self.ca_certs
                ssl_context_factory = None
                if self.insecure:
                    ssl_context_factory = gevent.ssl._create_unverified_context
                return httpclient.InferenceServerClient(
                    url=self.url,
                    concurrency=self.concurrency,
                    connection_timeout=self.connection_timeout,
                    network_timeout=self.network_timeout,
                    verbose=self.verbose,
                    ssl=True,
                    ssl_options=ssl_options,
                    insecure=self.insecure,
                    ssl_context_factory=ssl_context_factory,
                )
            else:
                return httpclient.InferenceServerClient(
                    url=self.url, verbose=self.verbose
                )
        except Exception as e:
            raise Exception("channel creation failed: " + str(e))

    @staticmethod
    def build_input(name: str, data: np.asarray, datatype: str = "INT32", binary_data=False):
        """
        创建input
        :param name: The name of input whose data will be described by this object
        :param data: numpy array, The tensor data in numpy array format
        :param datatype: np.dtype: INT8|INT16|INT32|INT64|UINT8|UINT16|UINT32|UINT64|FP16|FP32|FP64
        :param binary_data:
        :return:
        """
        infer_input = httpclient.InferInput(name, data.shape, datatype)
        infer_input.set_data_from_numpy(data, binary_data=binary_data)
        return infer_input

    @staticmethod
    def build_output(name: str, binary_data=False):
        """
        创建output
        :param name: The name of output
        :param binary_data:
        :return:
        """
        return httpclient.InferRequestedOutput(name, binary_data=binary_data)

    def infer_sync(self, inputs: list, model_name: str, model_version: str = "", http_headers: dict = None,
                   request_compression_algorithm: str = None, response_compression_algorithm: str = None, outputs: list = None):
        """
        同步推理
        :param inputs: A list of InferInput objects, each describing data for a input tensor required by the model.
        :param outputs: A list of InferRequestedOutput objects, each describing how the output data must be returned.
               If not specified all outputs produced by the model will be returned using default settings.
        :param model_name: The name of the model to run inference.
        :param model_version: The version of the model to run inference.
        :param model_version:
        :param http_headers:
        :param request_compression_algorithm:  Optional HTTP compression algorithm to use for the request body on client side.
               Currently supports "deflate", "gzip" and None. By default, no compression is used.
        :param response_compression_algorithm: Optional HTTP compression algorithm to request for the response body.
               Note that the response may not be compressed if the server does not
               support the specified algorithm. Currently supports "deflate",
               "gzip" and None. By default, no compression is requested.
        :return:
        """
        try:
            results = self.client.infer(
                model_name=model_name,
                model_version=model_version,
                inputs=inputs,
                outputs=outputs,
                headers=http_headers,
                request_compression_algorithm=request_compression_algorithm,
                response_compression_algorithm=response_compression_algorithm,
            )
            return results
        except InferenceServerException as e:
            print(e)

    def infer_async(self, inputs: list, model_name: str, model_version: str = "", http_headers: dict = None, request_compression_algorithm: str = None, response_compression_algorithm: str = None, outputs: list = None):
        return self.client.async_infer(
            model_name=model_name,
            model_version=model_version,
            inputs=inputs,
            outputs=outputs,
            headers=http_headers,
            request_compression_algorithm=request_compression_algorithm,
            response_compression_algorithm=response_compression_algorithm,
        )

    def infer_sync_demo(self, data, infer_count=20):
        """
        同步推理demo
        :param data:
        :param infer_count:
        :return:
        """
        start = time.time()
        for i in range(0, infer_count):
            start_time = time.perf_counter()
            result = self.infer_sync(model_name="cosnet_onnx", model_version="1",
                                     inputs=[client.build_input(name="input_ids", data=data[0], datatype="INT64"),
                                             client.build_input(name="token_type_ids", data=data[1], datatype="INT64"),
                                             client.build_input(name="attention_mask", data=data[2], datatype="INT64")],
                                     outputs=[client.build_output(name="output", binary_data=True)])
            end_time = time.perf_counter()
            run_time = end_time - start_time
            print(f"单次预测的时间：{run_time}秒")
        print("平均推理耗时：{0:.5f}".format((time.time()-start)/infer_count))
        # start = time.time()
        # for i in range(0, infer_count):
        #     start_time = time.perf_counter()
        #     result = self.infer_sync(model_name="cosnet_onnx", model_version="1",
        #                              inputs=[client.build_input(name="input_ids", data=data[0], datatype="INT64"),
        #                                      client.build_input(name="token_type_ids", data=data[1], datatype="INT64"),
        #                                      client.build_input(name="attention_mask", data=data[2], datatype="INT64")],
        #                              outputs=[client.build_output(name="output", binary_data=True)])
        #     end_time = time.perf_counter()
        #     run_time = end_time - start_time
        #     print(f"单次预测的时间：{run_time}秒")
        # print("平均推理耗时：{0:.5f}".format((time.time()-start)/infer_count))

    def infer_async_demo(self, data, infer_count=20):
        """
        异步推理demo
        :param data:
        :param infer_count:
        :return:
        """
        async_requests = []
        start = time.time()
        for i in range(infer_count):
            async_requests.append(
                self.infer_async(model_name="cosnet_onnx", model_version="1",
                                 inputs=[client.build_input(name="input_ids", data=data[0], datatype="INT64"),
                                         client.build_input(name="token_type_ids", data=data[1], datatype="INT64"),
                                         client.build_input(name="attention_mask", data=data[2], datatype="INT64")],
                                 outputs=[client.build_output(name="output", binary_data=True)])
            )
        res_output = []
        for async_request in async_requests:
            start_time = time.perf_counter()
            # Get the result from the initiated asynchronous inference request.
            # Note the call will block till the server responds.
            res_output.append(async_request.get_result())
            end_time = time.perf_counter()
            run_time = end_time - start_time
            print(f"单次预测的时间：{run_time}秒")
        print("平均推理耗时：{0:.20f}".format((time.time()-start)/infer_count))
        # start = time.time()
        # async_requests = []
        # for i in range(infer_count):
        #     async_requests.append(
        #         self.infer_async(model_name="cosnet_tensorrt", model_version="1",
        #                          inputs=[client.build_input(name="input_ids", data=data[0], datatype="INT32"),
        #                                  client.build_input(name="token_type_ids", data=data[1], datatype="INT32"),
        #                                  client.build_input(name="attention_mask", data=data[2], datatype="INT32")],
        #                          outputs=[client.build_output(name="output", binary_data=True)])
        #     )
        # res_output = []
        # for async_request in async_requests:
        #     start_time = time.perf_counter()
        #     # Get the result from the initiated asynchronous inference request.
        #     # Note the call will block till the server responds.
        #     res_output.append(async_request.get_result())
        #     end_time = time.perf_counter()
        #     run_time = end_time - start_time
        #     print(f"单次预测的时间：{run_time}秒")
        # print("平均推理耗时：{0:.20f}".format((time.time()-start)/infer_count))


if __name__ == "__main__":
    from zhousflib.ann import to_numpy
    from zhousflib.ann.torch.torch_to_onnx import example_inputs_demo
    args = example_inputs_demo(input_size=1)
    data_arr = np.asarray([to_numpy(args[0].int()), to_numpy(args[1].int()), to_numpy(args[2].int())], dtype=np.int64)
    client = ClientHttp(url="127.0.0.1:5005", concurrency=100)
    """
    同步请求demo
    """
    client.infer_sync_demo(data=data_arr, infer_count=500)
    """
    异步请求demo
    """
    # client.infer_async_demo(data=data_arr, infer_count=10)
    pass



# -*- coding: utf-8 -*-
# @Author  : zhousf
# @Date    : 2023/11/27 
# @Function: 人工神经网络

"""
############## 【安装CUDA】 ##############
CUDA下载：https://developer.nvidia.com/cuda-toolkit-archive
推荐版本：CUDA Toolkit 11.7
【windows】
新建系统环境变量：CUDA_PATH=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7
新建系统环境变量：CUDA_PATH_V11_7=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7
验证：在cmd中运行 nvcc -V
【linux】
vim ~/.bashrc
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/usr/local/cuda/lib64
export PATH=$PATH:/usr/local/cuda/bin
export CUDA_HOME=$CUDA_HOME:/usr/local/cuda
source ~/.bashrc
验证：
nvcc -V


############## 【安装cuDNN】 ##############
cudnn下载：https://developer.nvidia.com/rdp/cudnn-archive
推荐版本：cuDNN v8.9.0 for CUDA11.x
【windows】
将目录下的bin、lib（lib选择x64）、include文件夹复制到C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/11.7目录中
添加系统环境变量：Path=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/bin
添加系统环境变量：Path=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/libnvvp
验证：
在cmd中运行 C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/extras/demo_suite/deviceQuery.exe
显示Result = PASS则成功


############## 【安装tensorRT】 ##############
官方文档：https://docs.nvidia.com/deeplearning/tensorrt/archives/tensorrt-861/quick-start-guide/index.html
tensorrt下载：https://developer.nvidia.com/nvidia-tensorrt-8x-download
【windows】
将目录下的bin、lib（lib选择x64）、include文件夹复制到C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/11.7目录中
添加系统环境变量：Path=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/bin
添加系统环境变量：Path=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/libnvvp
添加系统环境变量：Path=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/include
添加系统环境变量：Path=C:/Program Files/NVIDIA GPU Computing Toolkit/CUDA/v11.7/lib/x64
设置完后重启电脑
python -m pip install TensorRT-8.6.1.6/python/tensorrt-8.6.1-cp39-none-linux_x86_64.whl
pip install TensorRT-8.6.1.6/uff/uff-0.6.9-py2.py3-none-any.whl
pip install TensorRT-8.6.1.6/graphsurgeon/graphsurgeon-0.4.6-py2.py3-none-any.whl
pip install TensorRT-8.6.1.6/onnx_graphsurgeon/onnx_graphsurgeon-0.3.12-py2.py3-none-any.whl
【linux】
下载地址：https://developer.nvidia.com/nvidia-tensorrt-download
1、选择deb file方式安装：
os="ubuntuxx04"
tag="8.x.x-cuda-x.x"
sudo dpkg -i nv-tensorrt-local-repo-${os}-${tag}_1.0-1_amd64.deb
sudo cp /var/nv-tensorrt-local-repo-${os}-${tag}/*-keyring.gpg /usr/share/keyrings/
sudo apt-get update
sudo apt-get install tensorrt
2、选择tar file方式安装：
解压：tar -zxvf TensorRT-8.6.1.6.Linux.x86_64-gnu.cuda-11.8.tar.gz
cd TensorRT-8.6.1.6/python
python -m pip install tensorrt-8.6.1-cp39-none-linux_x86_64.whl
配置环境变量：
建议使用(避免每次source)：vim /etc/profile
vim ~/.bashrc
export PATH=/root/zhousf/tensorrt/TensorRT-8.6.1.6/bin:$PATH
export LIBRARY_PATH=/root/zhousf/tensorrt/TensorRT-8.6.1.6/lib:$LIBRARY_PATH
export LD_LIBRARY_PATH=/root/zhousf/tensorrt/TensorRT-8.6.1.6/targets/x86_64-linux-gnu/lib:$LD_LIBRARY_PATH
source ~/.bashrc
建议使用：source /etc/profile
验证：
import tensorrt as trt
print(trt.__version__)


############## 【安装pycuda】 ##############
先安装cuda
安装文档：https://wiki.tiker.net/PyCuda/Installation/
python -m pip install pycuda==2022.1
若安装失败可以采用conda安装：conda install pycuda
若windows安装失败可以采用下载文件方式：https://www.lfd.uci.edu/~gohlke/pythonlibs/#pycuda
下面是下载源码编译方式：https://pypi.org/project/pycuda/#files
tar -zxvf pycuda-2023.1.tar.gz
cd pycuda-2023.1
python setup.py build
python setup.py install
验证：
import pycuda.driver as cuda
import pycuda.autoinit
cuda.init()
print("CUDA device count:", cuda.Device.count())

报错：PyCUDA The context stack was not empty upon module cleanup 解决方法：import pycuda.autoinit
报错：pycuda._driver.LogicError: cuMemcpyHtoDAsync failed: invalid argument  解决方法：数据没有格式化为int类型，to_numpy(data.int())
当预测结果都是0时，可能是trt多线程的问题
解决方法： 
    import pycuda.autoinit
    self.cfx = cuda.Device(0).make_context()
    self.cfx.push()
    推理代码...
    self.cfx.pop()


############## 【安装torch】 ##############
选择版本：https://pytorch.org/get-started/locally/
【cpu】
pip install torch==1.13.1+cpu torchvision==0.14.1+cpu torchaudio==0.13.1 --extra-index-url https://download.pytorch.org/whl/cpu
【gpu】
pip install torch==1.13.1+cu117 torchvision==0.14.1+cu117 torchaudio==0.13.1 --extra-index-url https://download.pytorch.org/whl/cu117
验证：
import torch
print(torch.__version__)


############## 【安装transformers(from HuggingFace)】 ##############
# 注意版本要一致，不然会报错：Unexpected key(s) in state_dict: "bert.embeddings.position_ids".
pip install transformers==4.30.2 -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com


【onnx && cuda的版本对应关系】
onnx对应cuda的版本：https://onnxruntime.ai/docs/execution-providers/CUDA-ExecutionProvider.html#requirements
注意onnxruntime与opset版本的对应关系


############## 【安装onnxruntime】 ##############
选择版本：https://onnxruntime.ai/docs/execution-providers/CUDA-ExecutionProvider.html#requirements
【cpu】
pip install onnxruntime==1.13.1 -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
【gpu】
pip install onnxruntime-gpu==1.13.1 -i http://mirrors.aliyun.com/pypi/simple/ --trusted-host mirrors.aliyun.com
验证：
import onnxruntime
onnxruntime.get_device()


############## 【验证导出onnx是否正确】 ##############
可视化网络结构：https://netron.app/
当output有if条件则会存在问题，更换opset版本(opset=10)或降低torch版本(1.8.0)


"""


def to_numpy(tensor):
    return tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()


def check_device_id(device_id: int = -1):
    """
    :param device_id: cpu上运行：-1 | gpu上运行：0 or 1 or 2...
    :return:
    """
    assert device_id >= -1, 'Expected device_id >= 1, but device_id={0}.'.format(device_id)

